# Copyright 2023 Eurobios Mews Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Analytical root finding with array inputs (test)."""

import numpy as np
import pytest

from pyntb.polynomial import solve_p2_v, _cardan_v, solve_p3_v

_nprs = 3141592654


def _maxabs(x):
    return np.max(np.abs(x))


@pytest.fixture
def sizetol():
    return 99, 1.0e-10


def test_solve_p2_distinct_real_roots(sizetol):
    np.random.seed(_nprs)
    size, tol = sizetol

    r1 = np.random.randn(size)
    r2 = np.random.randn(size)

    def fun(x):
        return (x - r1) * (x - r2)

    x1, x2 = solve_p2_v(1.0, -(r1 + r2), r1 * r2)

    assert np.logical_and.reduce(
        (
            _maxabs(fun(x1)) < tol,
            _maxabs(fun(x2)) < tol,
            np.max(np.minimum(np.abs(r1 - x1), np.abs(r1 - x2))) < tol,
            np.max(np.minimum(np.abs(r2 - x1), np.abs(r2 - x2))) < tol,
        )
    )


def test_solve_p2_double_real_root(sizetol):
    np.random.seed(_nprs)
    size, tol = sizetol

    rd = np.random.randn(size)

    def fun(x):
        return (x - rd) ** 2

    x1, x2 = solve_p2_v(1.0, -2 * rd, rd**2)

    assert np.logical_and.reduce(
        (
            _maxabs(fun(x1)) < tol,
            _maxabs(fun(x2)) < tol,
            _maxabs(x1 - rd) < tol,
            _maxabs(x2 - rd) < tol,
        )
    )


def test_solve_p2_complex_roots(sizetol):
    np.random.seed(_nprs)
    size, tol = sizetol

    r1 = np.random.randn(size) + 1j * np.random.randn(size)
    r2 = np.random.randn(size) + 1j * np.random.randn(size)

    def fun(x):
        return (x - r1) * (x - r2)

    x1, x2 = solve_p2_v(1.0, -(r1 + r2), r1 * r2)

    assert np.logical_and.reduce(
        (
            _maxabs(fun(x1)) < tol,
            _maxabs(fun(x2)) < tol,
            np.max(np.minimum(np.abs(r1 - x1), np.abs(r1 - x2))) < tol,
        )
    )


def test_solve_cardan(sizetol):
    np.random.seed(_nprs)
    size, tol = sizetol

    p = np.random.randn(size)
    q = np.random.randn(size)

    def fun(x):
        return x**3 + p * x + q

    x1, x2, x3 = _cardan_v(p, q)

    assert np.logical_and.reduce(
        (
            _maxabs(fun(x1)) < tol,
            _maxabs(fun(x2)) < tol,
            _maxabs(fun(x3)) < tol,
        )
    )


def test_solve_p3(sizetol):
    np.random.seed(_nprs)
    size, tol = sizetol

    a = np.random.randn(size)
    b = np.random.randn(size)
    c = np.random.randn(size)
    d = np.random.randn(size)

    def fun(x):
        return a * x**3 + b * x**2 + c * x + d

    x1, x2, x3 = solve_p3_v(a, b, c, d)

    assert np.logical_and.reduce(
        (
            _maxabs(fun(x1)) < tol,
            _maxabs(fun(x2)) < tol,
            _maxabs(fun(x3)) < tol,
        )
    )



# Copyright 2023 Eurobios Mews Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Geodata utility functions (test)."""

import numpy as np
from sklearn.metrics.pairwise import haversine_distances

from pyntb.geoutils import haversine

_nprs = 3141592654


def test_haversine():
    np.random.seed(_nprs)
    size = 99
    tol = 1.0e-12

    lat1 = -90 + 180 * np.random.rand(size)
    lat2 = -90 + 180 * np.random.rand(size)
    lon1 = -180 + 360 * np.random.rand(size)
    lon2 = -180 + 360 * np.random.rand(size)

    x = np.deg2rad(np.column_stack((lat1, lon1)))
    y = np.deg2rad(np.column_stack((lat2, lon2)))

    da = np.diag(haversine_distances(x, y))
    db = haversine(lat1, lon1, lat2, lon2, r=1)

    assert np.max(np.abs(db - da) <= tol)



# Copyright 2023 Eurobios Mews Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Analytical root finding with array inputs."""

from __future__ import annotations  # Type annotations for Python 3.7 and 3.8

from typing import Union

import numpy as np


def solve_p2_v(
    a: Union[float, np.ndarray],
    b: Union[float, np.ndarray],
    c: Union[float, np.ndarray],
) -> tuple[Union[float, np.ndarray], Union[float, np.ndarray]]:
    """
    Roots of second order polynomial with array as coefficients.

    Solve the equation a*x**2 + b*x + c = 0 when coefficients are arrays. If
    arrays are used, they must have the same size. Coefficient a must not be
    zero.

    Parameters
    ----------
    a :
    b :
    c :

    Returns
    -------
    x1: first root (may be complex)
    x2: second root (may be complex)

    """
    delta = b**2 - 4.0 * a * c
    i2a = 0.5 / a
    sqd = np.sqrt(delta + 0j)
    x1 = (-b + sqd) * i2a
    x2 = (-b - sqd) * i2a

    if np.all(np.imag(x1) == 0.0):
        x1 = np.real(x1)

    if np.all(np.imag(x2) == 0.0):
        x2 = np.real(x2)

    return x1, x2


def _cardan_v(
    p_: Union[float, np.ndarray], q_: Union[float, np.ndarray]
) -> tuple[
    Union[float, np.ndarray], Union[float, np.ndarray], Union[float, np.ndarray]
]:
    """
    Roots of specific third order polynomial with array as coefficients.

    Solve the equation x**3 + p*x + q = 0 when coefficients are arrays. If
    arrays are used, they must have the same size.

    Parameters
    ----------
    p_ :
    q_ :

    Returns
    -------
    x1: first root (may be complex)
    x2: second root (may be complex)
    x3: third root (may be complex)

    """
    p = np.asarray(p_)
    q = np.asarray(q_)

    delta = 4.0 * p**3 + 27.0 * q**2
    sp = np.sqrt(delta / 27.0)
    sm = np.sqrt(delta / 27.0 + 0j)

    u = np.where(
        delta >= 0.0, np.cbrt(0.5 * (-q + sp)), np.power(0.5 * (-q + sm), 1 / 3)
    )

    v = np.where(
        delta >= 0.0, np.cbrt(0.5 * (-q - sp)), np.power(0.5 * (-q - sm), 1 / 3)
    )

    j1 = 0.5 * (-1.0 + np.sqrt(3) * 1j)
    j2 = j1.conjugate()
    x1 = u + v
    x2 = j1 * u + j2 * v
    x3 = j2 * u + j1 * v

    return np.real(x1), x2, x3


def solve_p3_v(
    a: Union[float, np.ndarray],
    b: Union[float, np.ndarray],
    c: Union[float, np.ndarray],
    d: Union[float, np.ndarray],
) -> tuple[
    Union[float, np.ndarray], Union[float, np.ndarray], Union[float, np.ndarray]
]:
    """
    Roots of second order polynomial with array as coefficients.

    Solve the equation a*x**3 + b*x**2 + c*x + d = 0 when coefficients are
    arrays. If arrays are used, they must have the same size. Coefficient a must
    not be zero.

    Parameters
    ----------
    a :
    b :
    c :
    d :

    Returns
    -------
    x1: first root (may be complex)
    x2: second root (may be complex)
    x3: third root (may be complex)

    """
    p = -((b / a) ** 2) / 3.0 + c / a
    q = (b / (27.0 * a)) * (2.0 * (b / a) ** 2 - 9.0 * c / a) + d / a
    z1, z2, z3 = _cardan_v(p, q)

    s = b / (3.0 * a)
    x1 = z1 - s
    x2 = z2 - s
    x3 = z3 - s

    return x1, x2, x3



# Copyright 2023 Eurobios Mews Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Misc. numeric functions (test)."""

import numpy as np

from pyntb.optimize import bisect_v, fixed_point, qnewt2d_v

_nprs = 3141592654


def test_bisect():
    np.random.seed(_nprs)
    size = 99
    tol = 1.0e-09

    c = np.linspace(1, size, size) + np.random.randn(size)
    c = np.abs(c)

    def fun(x):
        return x**2 - c

    r0 = np.sqrt(c)

    x0, err = bisect_v(
        fun,
        0.0,
        np.sqrt(np.max(c)) * 1.1,
        (size,),
        print_err=False,
        tol=tol,
        maxiter=99,
    )

    assert np.max(np.abs(x0 - r0) <= np.minimum(tol, err))


def test_bisect_scalar():
    def f(x):
        return x**2 - 2

    tol = 1e-6
    x0, err = bisect_v(f, a=0, b=2, shape=1, tol=tol)
    np.testing.assert_allclose(x0, np.sqrt(2), atol=tol)


def test_bisect_vector():
    def f(x):
        return np.array([x[0] ** 2 - 2, x[1] ** 3 - 2])
        # Not the best way to use bisect_v, but just for testing.

    tol = 1e-6
    x0, err = bisect_v(f, a=0, b=2, shape=(2,), tol=tol)
    np.testing.assert_allclose(x0, np.array([np.sqrt(2), np.cbrt(2)]), atol=tol)


def test_bisect_array():
    c = np.arange(27).reshape(3, 3, 3)

    def f(x):
        return x**2 - c

    tol = 1e-6
    x0, err = bisect_v(f, a=0, b=30, shape=c.shape, tol=tol)
    np.testing.assert_allclose(x0, np.sqrt(c), atol=tol)


def test_fixed_point():
    size = 99
    tol = 1.0e-09

    c = 1 + np.sqrt(np.linspace(1, 10, size))

    def fun(x):
        return np.sin(c * x)

    x0 = fixed_point(fun, 0.5, xtol=tol, maxiter=999)

    assert np.max(np.abs(x0 - fun(x0))) < tol


def test_qnewt2d():
    np.random.seed(_nprs)
    size = 99
    tol = 1.0e-12

    a = np.abs(1 + np.random.randn(size))
    b = np.abs(1 + np.random.randn(size))

    def f1(x, y):
        return y - a * x**2

    def f2(x, y):
        return y - b * x**3

    xg = np.ones((size,))
    yg = np.ones((size,))
    x, y, count, err = qnewt2d_v(
        f1, f2, xg, yg, rtol=tol, maxiter=999, dx=1.0e-09, dy=1.0e-09
    )

    assert np.logical_and(
        np.max(np.abs(f1(x, y))) < tol, np.max(np.abs(f2(x, y))) < tol
    )



# Copyright 2023 Eurobios Mews Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Misc. numeric functions."""

import warnings

import numpy as np
from scipy._lib._util import _asarray_validated
from scipy._lib.array_api_extra import apply_where


def bisect_v(
    fun: callable,
    a: float,
    b: float,
    shape: tuple[int, ...],
    tol=1.0e-06,
    maxiter=128,
    print_err=False,
) -> tuple[np.ndarray, np.ndarray]:
    """Bisection method to find a zero of a continuous function [a, b] -> R,
    such that f(a) < 0 < f(b).

    The method is vectorized, in the sense that it can in a single call find
    the zeros of several independent real-valued functions with the same input
    range [a, b].
    For this purpose, the `fun` argument should be a Python function taking as
    input a Numpy array of values in [a, b] and returning a Numpy array
    containing the evaluation of each function for the corresponding input.
    This is most efficient if the outputs values of all these functions are
    computed using vectorized Numpy operations as in the example below.

    Parameters
    ----------
    fun: Python function taking a Numpy array and returning a Numpy array of the same shape.
    a : lower bound of [a, b] interval.
    b : upper bound of [a, b] interval.
    shape : shape of the inputs and ouputs of `fun` and thus of the ouputs of `bisect_v`.
    tol : absolute tolerance.
    maxiter : maximum number of iterations.
    print_err : print or not max error and number of iterations at the end.

    Returns
    -------
    x: Numpy array of the same shape as the inputs and outputs of `fun`
        The zero of each function.
    err: Numpy array of the same shape as the inputs and outputs of `fun`
        The convergence error.

    Example
    -------
    >>> c = np.array([1.0, 4.0, 9.0, 16.0])
    >>> def f(x):
    ...     return x**2 - c
    >>> x0, err = bisect_v(f, a=0.0, b=10.0, shape=(4,), tol=1e-10)
    >>> x0
    array([1., 2., 3., 4.])

    """
    a_ = a * np.ones(shape)
    b_ = b * np.ones(shape)

    err = np.abs(b - a)
    count = 1
    while np.nanmax(err) > tol and count <= maxiter:
        x = 0.5 * (a_ + b_)
        y = fun(x)
        i = y < 0
        a_[i] = x[i]
        b_[~i] = x[~i]
        err = np.abs(b_ - a_)
        count += 1
    x = 0.5 * (a_ + b_)
    x[np.isnan(fun(x))] = np.nan
    if print_err:
        print(f"Bisection max err (abs) : {np.max(err):.2E}; count={count}")
    return x, err


def _del2(p0, p1, d):
    return p0 - np.square(p1 - p0) / d


def _relerr(actual, desired):
    return (actual - desired) / desired


def _fixed_point_helper(func, x0, args, xtol, maxiter, use_accel):
    """Almost copied from scipy.optimize._minpack_py.py.

    Changed
        if np.all(np.abs(relerr) < xtol)
    into
        if np.nanmax(np.abs(relerr)) < xtol

    Misc correction to match pylint.

    """
    p0 = x0
    for _ in range(maxiter):
        p1 = func(p0, *args)
        if use_accel:
            p2 = func(p1, *args)
            d = p2 - 2.0 * p1 + p0
            p = apply_where(d != 0, (p0, p1, d), _del2, fill_value=p2)
        else:
            p = p1
        relerr = apply_where(p0 != 0, (p, p0), _relerr, fill_value=p)
        if np.nanmax(np.abs(relerr)) < xtol:
            return p
        p0 = p
    msg = f"Failed to converge after {maxiter} iterations, value is {p}"
    warnings.warn(msg)
    n = len(p)
    s = np.sum(np.abs(relerr) < xtol)
    warnings.warn(f"{s} out of {n} element(s) in input array successfuly converged")
    return np.where(np.abs(relerr) < xtol, p, np.nan)


def fixed_point(func, x0, args=(), xtol=1e-8, maxiter=500, method="del2"):
    """
    Find a fixed point of the function (copied from
    scipy.optimize._minpack_py.py in order to handle nans).

    Given a function of one or more variables and a starting point, find a
    fixed point of the function: i.e., where ``func(x0) == x0``.

    Parameters
    ----------
    func : callable
        Function to evaluate.
    x0 : array_like
        Fixed point of function.
    args : tuple, optional
        Extra arguments to `func`.
    xtol : float, optional
        Convergence tolerance, defaults to 1e-08.
    maxiter : int, optional
        Maximum number of iterations, defaults to 500.
    method : {"del2", "iteration"}, optional
        Method of finding the fixed-point, defaults to "del2",
        which uses Steffensen's Method with Aitken's ``Del^2``
        convergence acceleration [1]_. The "iteration" method simply iterates
        the function until convergence is detected, without attempting to
        accelerate the convergence.

    """
    use_accel = {"del2": True, "iteration": False}[method]
    x0 = _asarray_validated(x0, as_inexact=True)
    return _fixed_point_helper(func, x0, args, xtol, maxiter, use_accel)


def qnewt2d_v(
    f1: callable,
    f2: callable,
    x0: np.ndarray,
    y0: np.ndarray,
    rtol=1.0e-12,
    maxiter=64,
    dx=1.0e-03,
    dy=1.0e-03,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Two-dimensional quasi-Newton with arrays.

    Apply a 2D quasi newton on a large number of case at the same time, ie solve
    the system [f1(x, y), f2(x, y)] = [0., 0.] in n cases.

    Derivatives are estimated with a second-order centered estimation (ie f1 and
    f2 are evaluated four times at each iteration).

    All return values are arrays of the same size as inputs x0 and y0.

    Parameters
    ----------
    f1 : first component of a 2d function of two variables
    f2 : second component of a 2d function of two variables
    x0 : first component of the initial guess
    y0 : second component of the initial guess
    rtol : relative tolerance
    maxiter : max number of iteration
    dx : delta for evaluating derivative regarding first component
    dy : delta for evaluating derivative regarding second component

    Returns
    -------
    x: first component of solution
    y: second component of solution
    count: number of iterations when exiting the function
    err: relative error when exiting the function

    """
    err = 1.0
    count = 0
    x = x0.copy()
    y = y0.copy()
    while err > rtol and count < maxiter:
        F1 = f1(x, y)
        F2 = f2(x, y)
        Ja = 0.5 * (f1(x + dx, y) - f1(x - dx, y)) / dx
        Jb = 0.5 * (f1(x, y + dy) - f1(x, y - dy)) / dy
        Jc = 0.5 * (f2(x + dx, y) - f2(x - dx, y)) / dx
        Jd = 0.5 * (f2(x, y + dy) - f2(x, y - dy)) / dy
        di = 1.0 / (Ja * Jd - Jb * Jc)
        ex = di * (Jd * F1 - Jb * F2)
        ey = di * (Ja * F2 - Jc * F1)
        x -= ex
        y -= ey
        err = max(np.nanmax(np.abs(ex / x)), np.nanmax(np.abs(ey / y)))
        count += 1
    return x, y, count, np.maximum(np.abs(ex / x), np.abs(ey / y))



# Copyright 2023 Eurobios Mews Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Geodata utility functions."""

import numpy as np


def haversine(lat1, lon1, lat2, lon2, r=6.371e06):
    """Haversine distance.

    Estimate distance between to gps points using the haversine formula (see
    https://en.wikipedia.org/wiki/Haversine_formula). Note that since the Earth
    is not a perfect sphere, this formula is not exact, and tends to be more
    wrong near the poles. The default value is the Earth's mean radius given in
    meters, you can change it to any value to fit your problem.

    If inputs are numpy arrays, they must all have the same shape.

    Parameters
    ----------
    lat1 : float or numpy array
        Latitude of first point in degrees.
    lon1 : float or numpy array
        Longitude of first point in degrees.
    lat2 : float or numpy array
        Latitude of second point in degrees.
    lon2 : float or numpy array
        Longitude of second point in degrees.
    r : float
        Reference radius, ie the Earth's mean radius in meters in our case. The
        default is 6.371E+06.

    Returns
    -------
    Distance between input points.

    """
    p1 = np.radians(lat1)
    p2 = np.radians(lat2)
    dp = np.radians(lat2 - lat1)
    dl = np.radians(lon2 - lon1)
    a = np.sin(0.5 * dp) ** 2 + np.cos(p1) * np.cos(p2) * np.sin(0.5 * dl) ** 2
    c = np.arctan2(np.sqrt(a), np.sqrt(1 - a))
    return 2 * r * c
